/*=====================================================================

  File:      spHtmGeoCsharp.sql for Spatial Sample
  Summary:   Implements the Transact-SQL interface to the Spatial library
			 Adds/Replaces geographical lat/lon functions ONLY. Assumes, that
		     spHtmCsharp.sql has already been run
  Date:	     August 19, 2005

---------------------------------------------------------------------
  This spatial search library was developed by Alexander Szalay, Robert Brunner, 
  Peter Kunszt, and Gyorgy Fekete at the Department of Physics and Astronomy, 
  The Johns Hopkins University, in collaboration with Jim Gray, Microsoft Research.
  Details of the algorithms can be found at http://www.sdss.jhu.edu/htm/.
 ======================================================= */

--=================================================================================
-- spHtmCsharp.sql 
--   Jim Gray, Alex Szalay, Gyorgy Fekete, Mar 2006
-----------------------------------------------------------------------------------
-- Define load the Spatial Assembly (dll) and define the entry points to
-- Geographical (lat/lon based) the HTM functions.
--  
-- Defines: 
--		fHtmLatLon
--		fHtmXyzToLatLon
--      fDistanceLatLon
--		fHtmCoverCircleLatLon
--      fHtmNearbyLatLon		-- nearby objects
--      fHtmNearestLatLon		-- nearest object
----------------------------------------------------------------------------------
-- History: (see spHtmCsharp.sql)

----------------------------------------------------------------------------------
--* todo:   
----------------------------------------------------------------------------------
SET NOCOUNT ON
go
exec sp_configure 'clr enabled', 1
reconfigure with override
go
if exists (select * from sys.objects where [name] = N'fHtmLatLon') 
			drop function  fHtmLatLon
if exists (select * from sys.objects where [name] = N'fHtmXyzToLatLon') 
			drop function fHtmXyzToLatLon
if exists (select * from sys.objects where [name] = N'fHtmLatLonToXyz')
			drop function fHtmLatLonToXyz
if exists (select * from sys.objects where [name] = N'fDistanceLatLon') 
			drop function fDistanceLatLon
if exists (select * from sys.objects where [name] = N'fDistanceXyz') 
			drop function fDistanceXyz
if exists (select * from sys.objects where [name] = N'fHtmCoverCircleLatLon') 
			drop function  fHtmCoverCircleLatLon
if exists (select * from sys.objects where [name] = N'fHtmNearbyLatLon') 
			drop function fHtmNearbyLatLon
if exists (select * from sys.objects where [name] = N'fHtmNearestLatLon') 
			drop function fHtmNearestLatLon
--

------------------------------------------------------------------
/* release version */
-- Add the assembly which contains the CLR methods we want to invoke on the server.
DECLARE @SamplesPath nvarchar(1024)
-- You may need to modify the value of the this variable if you have installed the sample someplace other than the default location.
SELECT @SamplesPath = 'C:\apps\jhucas\bin\' 
--
GO
------------------------------------------------------------------
--/* debug version */
-- Add the assembly which contains the CLR methods we want to invoke on the server.

--ALTER ASSEMBLY Spatial 
--  ADD FILE FROM @SamplesPath + 'Debug\SphericalHTM.pdb';
--GO

------------------------------------------------------------------------
-- Now define the functions in the HTM assembly
-----------------------------------------------
go
create function fHtmLatLon(@Lat float, @Lon float) returns bigint
------------------------------------------------------------- 
--/H Returns Hierarchical Triangular Mesh (HTM) ID of a given point.  
--/H <br> given right ascencion and declination in J2000 reference frame
------------------------------------------------------------- 
--/T <li> Lat float NOT NULL,          --/D Latitude will be nomalized to [0..360] --/U deg --/K POS_LIMIT
--/T <li> Lon float NOT NULL,          --/D Longitude will be nomalized to [-90..90] --/U deg --/K POS_LIMIT
--/T <li> returns htmID bigint         --/D 21-deep hierarchical trangular mesh ID of this point --/K CODE_HTM
--/T <br> 
--/T Sample call    
--/T <br><samp> 
--/T <br> declare @htmID bigint
--/T <br> select @htmID = dbo.fHtmEq(0, 195) 
--/T </samp>  
--/T <br> see also fHtmXyz, fHtmEq
------------------------------------------------------------- 
	as external name Spatial.[Spherical.Htm.Sql].fHtmLatLon
go
create function fDistanceLatLon(@Lat1 float, @Lon1 float, @Lat2 float, @Lon2 float) 
-------------------------------------------------------------
--/H returns distance (arc minutes) between two points (Lat1, Lon1) and (Lat2, Lon2)
-------------------------------------------------------------
--/T <br> Lat1, Lon1, Lat2, Lon2 are in degrees
--/T <br>
-------------------------------------------------------------
    returns float
    as external name Spatial.[Spherical.Htm.Sql].fDistanceLatLon
go
create function fHtmLatLonToXyz(@ra float, @dec float)
------------------------------------------------------------------
--/H Convert Lat, Lon to Cartesian coordinates (x, y, z)
------------------------------------------------------------------
--/T <br>Parameters:
--/T <li>Lat float --D/ Latitude in degrees
--/T <li>Lon float --D/ Longitude in degrees
--/T <li> returns Table of one row containing the values (x, y, z)
--/T <samp>select * from fHtmLatLonToXyz(0.0, 270.0)
--/T <br> gives:  x  y  z
--/T <br>         0 -1  0
--/T </samp>
--/T <br>
-----------------------------------------------------------------
				returns table (x float, y float, z float)
as external name Spatial.[Spherical.Htm.Sql].fHtmLatLonToXyz
go
create function fHtmXyzToLatLon(@x float, @y float, @z float)
------------------------------------------------------------------
--/H Convert Cartesian coordinates (x, y, z) to Lat/Lon pair.
--/H <br> (x, y, z) will be normalized unless (x, y, z) is close to (0,0,0)
--/H <br> 
------------------------------------------------------------------
--/T <br>Parameters:
--/T <li>x float --D/ 
--/T <li>y float --D/ 
--/T <li>z float --D/ 
--/T <li> returns Table of one row containing the values (Lat, Lon)
--/T <samp>select * from fHtmXyzToLatLon(0.0, 0.0, -1.0)
--/T <br> gives:  Lat Lon
--/T <br>         -90   0
--/T </samp>
--/T <br>
-----------------------------------------------------------------
				returns table (Lat float, Lon float)
as external name Spatial.[Spherical.Htm.Sql].fHtmXyzToLatLon
go
create function fHtmCoverCircleLatLon (@Lat float, @Lon float, @radiusArcMinutes float) 
	           returns table  (HtmIDStart bigint, HtmIDEnd bigint)
------------------------------------------------------------- 
--/H Returns HTM TrixelTable covering a circle centered at J2000 Lon,Lat (in degrees)
--/H <br> within  @radiusArcMinutes arcminutes of that centerpoint.
------------------------------------------------------------- 
--/T <li> Lat float NOT NULL,          --/D Latitude will be nomalized to [-90..90] --/U deg --/K POS_LIMIT
--/T <li> Lon float NOT NULL,         --/D Longitude will be nomalized to [0..360] --/U deg --/K POS_LIMIT
--/T <li> radiusArcMinutes float NOT NULL,  --/D radius in arcminutes --/U arcmin --/K POS_LIMIT
--/T <li> returns trixel table(HtmIDStart bigint, HtmIDEnd bigint)
--/T <br>	HtmIDStart bigint        --/D 21-deep HtmID range start --/K CODE_HTM
--/T <br>	HtmIDEnd   bigint        --/D 21-deep HtmID range end --/K CODE_HTM
--/T <br> 
--/T Sample call    
--/T <br><samp> 
--/T <br> select * from fHtmCoverCircleEq( 0, 190, 1)
--/T </samp>  
--/T <br> see also fHtmCoverCircleXyz, fHtmCover
------------------------------------------------------------- 
    as external name Spatial.[Spherical.Htm.Sql].fHtmCoverCircleLatLon 
-- select * from fHtmCoverCircleEq( 0, 190,1)
go
CREATE FUNCTION fHtmNearbyLatLon(@type char(1), @Lat float, @Lon float, @r float)
-------------------------------------------------------------
--/H Returns table of objects of the given type within @r arcmins of an xyz point (@Lat,@Lon).
-------------------------------------------------------------
--/T <li> type char(1) NOT NULL, --/D type of station: either 'P' for place or 'S' for stream flow station
--/T <li> Lat float NOT NULL,    --/D Latitude of center point  
--/T <li> Lon float NOT NULL,    --/D Longitude of center point 
--/T <li> r float NOT NULL,      --/D radius in arcminutes --/U arcmin --/K POS_LIMIT
--/T There is no limit on the number of objects returned.
--/T <p>returned table has the same format as the spatail index (minus the type field):  
--/T <li> htmID bigint,               -- Hierarchical Trangular Mesh id of this object
--/T <li> Lat, Lon float not null,    -- Latitude and Longitude (dec/ra) of point.   
--/T <li> x,y,z float not null,       -- x,y,z of unit vector to this object
--/T <li> objID bigint,               -- object ID in SpatialIndex table. 
--/T <li> distance float              -- distance in arc minutes to this object from the ra,dec.
--/T <br> Sample call to find places within 50 arcminutes of SQL Server Development. 
--/T <br><samp>
--/T <br>select *
--/T <br> from fHtmNearbyLatLon('P',47.5,-122) 
--/T </samp>  
--/T <br>see also fHtmNearbyEq, fHtmNearbyXYZ, fHtmNearestLatLon
-------------------------------------------------------------
  returns @SpatialIndex table (
					HtmID	bigint NOT NULL,
					Lat		float  NOT NULL,
					Lon		float  NOT NULL,
					x		float  NOT NULL,
					y		float  NOT NULL,
					z		float  NOT NULL,
 					ObjID	bigint NOT NULL,
					distance float NOT NULL -- distance in arc minutes
  ) as begin
	if (@r<0) RETURN
	-- compute x,y,z to make distance test fast.
	declare @x float, @y float, @z float
	select	@x = cos(radians(@Lat))*cos(radians(@Lon)),
			@y = cos(radians(@Lat))*sin(radians(@Lon)),
			@z = sin(radians(@Lat))
	-- do the spatial join using the HTM cover. 
	insert @SpatialIndex select 
	    HtmID, 
	    Lat,Lon,
	    x,y,z,
	    ObjID,
 	    2*degrees(asin(sqrt(power(@x-x,2)+power(@y-y,2)+power(@z-z,2))/2))*60 
	    --sqrt(power(@x-x,2)+power(@y-y,2)+power(@z-z,2))/PI()/3 
	    from fHtmCoverCircleXyz(@x,@y,@z, @r) join SpatialIndex 
			on HtmID BETWEEN  HtmIDStart AND HtmIDEnd  
            and [Type]  = @type
--          this clause is simplified since it is innner loop.
--	        and( (2*DEGREES(ASIN(sqrt(power(@x-x,2)+power(@y-y,2)+power(@z-z,2))/2))*60)< @r)
			and (power(@x-x,2)+power(@y-y,2)+power(@z-z,2)) < power(2*sin(radians(@r/120)),2) 
	order by (2*degrees(asin(sqrt(power(@x-x,2)+power(@y-y,2)+power(@z-z,2))/2))*60) asc
	OPTION(FORCE ORDER, LOOP JOIN)
	return
	end
go
CREATE FUNCTION fHtmNearestLatLon(@type char(1), @Lat float, @Lon float)
-------------------------------------------------------------
--/H Returns table of objects of the given type within @r arcmins of a ra/dec point.
-------------------------------------------------------------
--/T <li> Lat float NOT NULL,          --/D Latitude (decimal degrees) 
--/T <li> Lon float NOT NULL,         --/D Longitude (decimal degrees) 
--/T <p> One object is returned. 
--/T <br>returned table has the same format as the spatail index (minus the type field):  
--/T <li> htmID bigint,               -- Hierarchical Trangular Mesh id of this object
--/T <li> Lat, Lon float not null,    -- Latitude and Longitude (dec/ra) of point.   
--/T <li> x,y,z float not null,       -- x,y,z of unit vector to this object
--/T <li> objID bigint,               -- object ID in SpatialIndex table. 
--/T <li> distance float              -- distance in arc minutes to this object from the ra,dec.
--/T <br> Sample call to find nearest place to Baltimore (which is Baltimore (!)). 
--/T <br><samp>
--/T <br> select distance, Place.*
--/T <br> from fHtmNearestLatLon('P', 39.3, -76.6) I join Place on I.objID = Place.HtmID
--/T </samp>  
--/T <br>see also fHtmNearbyLatLon, fHtmNearbyEq, fHtmNearestXYZ
-------------------------------------------------------------
  returns @SpatialIndex table (
					HtmID	bigint NOT NULL,
					Lat		float  NOT NULL,
					Lon		float  NOT NULL,
					x		float  NOT NULL,
					y		float  NOT NULL,
					z		float  NOT NULL,
 					ObjID	bigint NOT NULL,
					distance float NOT NULL -- distance in arc minutes
  ) as begin
	declare @x float, @y float, @z float, @r  float
	select	@x = cos(radians(@Lat))*cos(radians(@Lon)),
			@y = cos(radians(@Lat))*sin(radians(@Lon)),
			@z = sin(radians(@Lat)),	
			@r = 1
-- Try r = 1, 4, 14,.... till you find a non null set.
-- do the spatial join using the HTM cover. 
retry:
	insert @SpatialIndex 
		select top 1 
	    HtmID, 
	    Lat,Lon,
	    x,y,z,
	    ObjID,
 	    2*degrees(asin(sqrt(power(@x-x,2)+power(@y-y,2)+power(@z-z,2))/2))*60 
	    --sqrt(power(@x-x,2)+power(@y-y,2)+power(@z-z,2))/PI()/3 
	    from fHtmCoverCircleXyz(@x,@y,@z, @r) join SpatialIndex 
			on HtmID BETWEEN  HtmIDStart AND HtmIDEnd  
            and [Type]  = @type
--          this clause is simplified since it is innner loop.
--	        and( (2*DEGREES(ASIN(sqrt(power(@x-x,2)+power(@y-y,2)+power(@z-z,2))/2))*60)< @r)
			and (power(@x-x,2)+power(@y-y,2)+power(@z-z,2)) < power(2*sin(radians(@r/120)),2) 
		order by (2*degrees(asin(sqrt(power(@x-x,2)+power(@y-y,2)+power(@z-z,2))/2))*60) asc
		OPTION(FORCE ORDER, LOOP JOIN)
		
	if (@@rowcount = 0)
		begin
		set @r = @r * 4
		goto retry
		end
	return
	end
go

/*
select dbo.fHtmEq(195,0)
select * from fHtmCover('CIRCLE J2000 195 0 1')
select * from fHtmCoverCircleEq(195,0,1)
select * from fHtmCover('REGION CARTESIAN CONVEX 1 0 0 -0.6 0 1 0 -0.6 0 0 1 -0.6 -1 0 0 -0.6 0 -1 0 -0.6 0 0 -1 -0.6')
select dbo.fHtmToNormalForm('REGION CARTESIAN CONVEX 1 0 0 -0.6 0 1 0 -0.6 0 0 1 -0.6 -1 0 0 -0.6 0 -1 0 -0.6 0 0 -1 -0.6')
*/
go
declare @version varchar(max)
select @version = dbo.fHtmVersion()
PRINT '[spHtmCsharp.sql]: Version: ' + @version + ' of the HTM library successfully appended with GEO.'
GO
